﻿using System.Collections;
using System.Threading;
using Unity.Jobs;
using UnityEngine;

/// <summary>
/// Divide the terrain generation into 9 "chunks".
/// Each chunk is responsible for the generation of its
/// terrain object and only its terrain object.
/// 
/// A chunk may be moved when the player moves, such that a 3x3
/// grid of chunks surrounds the player at any time, and that
/// there's a minimum of one full terrain between the player and
/// the void.
/// 
/// Once a chunk is moved, it will regenerate its terrain object.
/// This will be done by a thread/job.
/// </summary>
public class TerrainChunker : MonoBehaviour
{
    [Header("Autogenerated values")]
	[Tooltip("Current grid position of the chunker")]
    public Vector2Int gridPosition;

    public Vector2 crawlOffset;

    public TerrainMaster master;

    private Terrain terrain;

    private TerrainData data;

    private ShapeGeneratorJob shapeGenerator;
    private JobHandle runningShape;

    public void Initialize(TerrainMaster master, Vector2Int gridPosition,
        Vector2 crawlOffset)
    {
        this.master = master;
        this.gridPosition = gridPosition;
        this.crawlOffset = crawlOffset;

        gameObject.name = $"Chunker ({gridPosition})";

        data = Instantiate(Resources.Load<TerrainData>("Template"));
        terrain = gameObject.AddComponent<Terrain>();

        UpdateTerrain();
    }

    public void UpdatePositions(Vector2Int newGridPosition)
    {
        if (newGridPosition != gridPosition)
        {
            gridPosition = newGridPosition;

            // Now we update the terrain
            UpdateTerrain();
        }
    }

    private void UpdateTerrain()
    {
        if (!runningShape.IsCompleted)
        {
            // Cancel thread if its running.
            shapeGenerator.Cancel();
        }

        shapeGenerator = new(crawlOffset, gridPosition, 513, master.octaves);
        runningShape = shapeGenerator.Schedule();
    }

    private void ApplyTerrain(float[,] heights)
    {

    }

    private void Update()
    {
        if (shapeGenerator.Finished)
        {
            ApplyTerrain(shapeGenerator.Retrieve());
        }
    }

    private void OnDestroy()
    {
        shapeGenerator.Cancel();
    }
}