﻿using System.Collections;
using System.Collections.Generic;
using System.Linq;
using Unity.Collections;
using Unity.Jobs;
using Unity.Mathematics;
using UnityEngine;

public struct ShapeGeneratorJob : IJob
{
    #region Generation variables
    private Vector2 crawlOffset;

    private Vector2Int gridOffset;

    [ReadOnly]
    [DeallocateOnJobCompletion]
    private NativeArray<Octave> octaves;

    private int samples;
    #endregion
    #region Output variables
    private NativeArray<float> heights;
    #endregion
    #region Control variables
    public bool Finished { get; private set; }

    public bool Cancelled { get; private set; }
    #endregion

    public ShapeGeneratorJob(Vector2 crawlOffset, Vector2Int gridOffset,
        int samples, IEnumerable<Octave> octaves)
    {
        this.crawlOffset = crawlOffset;
        this.gridOffset = gridOffset;
        this.octaves = new(octaves.ToArray(), Allocator.TempJob);
        this.samples = samples;

        heights = new(samples, Allocator.Persistent);

        Finished = false;
        Cancelled = false;
    }

    #region Job control
    public void Execute()
    {
        Generate();
    }

    public void Cancel()
    {
        Cancelled = true;
        // Why the hell can I not deallocate them in a job.
        // octaves.Dispose();
        heights.Dispose();
    }

    public float[,] Retrieve()
    {
        if (!Finished)
            throw new System.ArgumentException("Not finished");
        else
        {
            float[,] arrHeights = new float[samples, samples];
            
            for (int y = 0; y < samples; y++)
            {
                for (int x = 0; x < samples; x++)
                {
                    arrHeights[y, x] = heights[y * samples + x];
                }
            }

            heights.Dispose();
            return arrHeights;
        }
    }
    #endregion

    public void Generate()
    {
        foreach (var octave in octaves)
        {
            for (int y = 0; y < samples; y++)
            {
                for (int x = 0; x < samples; x++)
                {
                    if (Cancelled)
                        return;

                    float2 crawlPos = new((y + crawlOffset.x * (samples - 1)) * octave.frequency,
                        (x + crawlOffset.y * (samples - 1)) * octave.frequency);
                    
                    float addition;

                    switch (octave.type)
                    {
                        default:
                        case NoiseType.MathfPerlin:
                            addition = Mathf.PerlinNoise(crawlPos.x, crawlPos.y);
                            break;
                        case NoiseType.MathematicsPerlin:
                            addition = noise.cnoise(crawlPos) / 2.3f + 0.5f;
                            break;
                        case NoiseType.Simplex:
                            addition = noise.snoise(crawlPos) / 4.6f + 0.5f;
                            break;
                        case NoiseType.Celluar:
                            addition = noise.cellular(crawlPos).x / 2.3f + 0.5f;
                            break;
                    }
                    heights[y * samples + x] += addition * octave.amplitude;
                }
            }
        }

        // octaves.Dispose();
        Finished = true;
    }
}